/* image.c - part of ziproxy package
 * 2002 Juraj Variny <variny@naex.sk>
 *
 * Released subject to GNU General Public License
 *
 * Image reading/writing wrappers.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include <gif_lib.h>

#include <jpeglib.h>
//remove some leaks from jpeglib headers:
#undef METHODDEF
#undef EXTERN
#undef LOCAL
#undef GLOBAL

#include <png.h>

#ifdef JP2K
#include <jasper/jasper.h>
#endif

#define SRC_IMAGE_C
#include "image.h"
#include "cfgfile.h"
//#include "log.h"


static int outlen;
static char *outbuf;

//Forwards. There are more utility functions, but they're used only once.
static raw_bitmap *new_raw_bitmap();

static int png2bitmap(char *inbuf, int insize, raw_bitmap **out);
static int gif2bitmap(char *inbuf, int insize, raw_bitmap **out);
static int jpg2bitmap(char *inbuf, int insize, raw_bitmap **out);

static int bitmap2jpg(raw_bitmap  * bmp, int quality, char ** outb, int * outl);

static int bitmap2png(raw_bitmap  * bmp, char ** outb, int * outl);

static int rgb2gray(raw_bitmap * bmp);
static int depalettize(raw_bitmap * bmp);

#ifdef JP2K
static int bitmap2jp2(raw_bitmap * bmp, float rate, char ** outb, int * outl);
#endif

int detect_type(char * line, int len){
	int image = OTHER_CONTENT;
	if(len <  4) return OTHER_CONTENT;
    // Check for image signatures
	if ((line[0] == -1) && (line[1] == -40) &&
		(line[2] == -1) && (line[3] == -32))
		image = IMG_JPEG;
	else if ((line[0] == 'G') && (line[1] == 'I') &&
		     (line[2] == 'F') && (line[3] == '8'))
		image = IMG_GIF;
	else if ((line[0] == -119) && (line[1] == 'P') &&
		     (line[2] == 'N') && (line[3] == 'G'))
		image = IMG_PNG;
	return image;
}

/* What will be done here with images:
 *
 * Compiled without JP2K, regardless of HTML modification
 * Input -> Output
 * GIF	-> PNG or JPEG, whatever is smaller
 * PNG	-> JPEG
 * JPEG	-> JPEG
 *
 * Compiled with JP2K, HTML modification on
 * Input -> Output if in OBJECT tag : Output if outside OBJECT tag
 * GIF	-> JP2K : PNG or JPEG, whatever is smaller
 * PNG	-> JP2K : JPEG
 * JPEG	-> JP2K : JPEG
 * 
 * Compiled with JP2K, HTML modification off
 * Input -> Output
 * GIF	-> PNG or JP2K, whatever is smaller
 * PNG	-> JP2K
 * JPEG	-> JP2K
 *
 * Compression GIF -> PNG is tried even if relevant ImageQuality or JP2Rate 
 * option is set to zero.
 */

int compress_image(http_headers *serv_hdr, http_headers *client_hdr, char *inbuf, int insize, char ** outb, int *outl){	
	int i, st = -1, pngstatus = -1, jp2status = -1, jpegstatus = -1, outtype = OTHER_CONTENT;
	int q, quality = -1;
	int try, TryPNG, TryJPEG, TryJP2;
	char *tempp;
	raw_bitmap *bmp;
	
	float rate = -1.0;

	try = detect_type(inbuf, insize);

	if (try != OTHER_CONTENT) serv_hdr->type = try;
	
	switch (serv_hdr->type) {
		case IMG_PNG:if(insize >= 600) st = png2bitmap(inbuf, insize, &bmp);
				     else st = -1;
			 quality = 128;
			 break;
		case IMG_GIF:if(insize >= 250) st = gif2bitmap(inbuf, insize, &bmp);
				     else st = -1;
			 quality = 128;
			 break;
		case IMG_JPEG:if(insize >= 600) st = jpg2bitmap(inbuf, insize, &bmp);
				      else st = -1;
			 quality = 0;
			 break;
					 
	}

	if (st) {//error, forward unchanged
		*outb = inbuf;
		*outl = insize;
		return st;
	}
	//worsest compression ratio allowed 85%(large images) - 90% (small)
	outlen = insize * 0.85 + 30;
	//What about this? Or make this configurable?
	//outlen = insize * 0.7;
	
	outbuf = (char *)malloc(outlen);

	tempp = outbuf;
	try = outlen;

	getRateQuality(bmp->width, bmp->height, &q, &rate);
	
#ifdef JP2K	
	TryPNG = ((IMG_GIF == serv_hdr->type) && !(client_hdr->flags & H_SUFF_MODIFIED) );
	TryJP2 = ((!ModifySuffixes  || (client_hdr->flags & H_SUFF_MODIFIED)) && (rate > 0.0) && (insize >= 600));
	TryJPEG = (!(client_hdr->flags & H_SUFF_MODIFIED) && (q != 0) && (insize >= 600));
#else
	TryPNG = ((IMG_GIF == serv_hdr->type));
	TryJP2 = 0;
	TryJPEG = ((!ModifySuffixes  || (client_hdr->flags & H_SUFF_MODIFIED)) && (q != 0) && (insize >= 600));
#endif
/*	logprintf("Will try compress image to: %s %s %s", 
			TryPNG ? "PNG" : "",
			TryJPEG ? "JPEG" : "",
			TryJP2 ? "JP2" : "");*/
	
	//Go compressing the image.

#ifdef JP2K
	if(TryJP2)
		if(rate*bmp->width*bmp->height*bmp->bpp > outlen) jp2status = 1;
		else{
			if(!bmp->bitmap) depalettize(bmp);
			jp2status = bitmap2jp2(bmp, rate, outb, outl);
		}
	
	else
		jp2status = 1;
#endif
	
	if(TryJPEG){
		if(!bmp->bitmap) depalettize(bmp);
		if (q < 0){
			q = -q;
			rgb2gray(bmp);
		}
		quality += q;

		jpegstatus = bitmap2jpg(bmp, quality, &outbuf, &outlen);
	}

	//Try PNG only if JPEG/JP2 didn't compressed it enough.
	try=0;
	if(TryPNG && (jp2status || (try=*outl) > insize/2) && 
			(jpegstatus || (try=outlen) > insize/2)){
		if(!try) try = insize * 0.85 + 30;
		tempp = (char*)malloc(try);
		pngstatus = bitmap2png(bmp, &tempp, &try);
	}

	//See what happened.
	if(!TryJP2 && (TryPNG || TryJPEG)){
		if(!jpegstatus && pngstatus){
			*outb = outbuf;
			*outl = outlen;
			outtype = IMG_JPEG;
		}else if(!pngstatus){
			*outb = tempp;
			*outl = try;
			outtype = IMG_PNG;
		}else{
			*outb = inbuf;
			*outl = insize;
			outtype = OTHER_CONTENT;
		}
	
	}else if (TryPNG || TryJP2){
		if( !jp2status && pngstatus){
			outtype = IMG_JP2K;
		} else if(!pngstatus){
			*outb = tempp;
			*outl = try;
			outtype = IMG_PNG;
		} else {
			*outb = inbuf;
			*outl = insize;
			outtype = OTHER_CONTENT;
		}
	
	}else assert(1);
	
	if (serv_hdr->where_content_type > 0){ 
		if(outtype != OTHER_CONTENT)
			switch(outtype){
				case IMG_JP2K:
					serv_hdr->hdr[serv_hdr->where_content_type] =
						"Content-Type: image/jp2";
					break;
				case IMG_JPEG:		
					serv_hdr->hdr[serv_hdr->where_content_type] = 
						"Content-Type: image/jpeg";
					break;
				case IMG_PNG:
					serv_hdr->hdr[serv_hdr->where_content_type] = 
						"Content-Type: image/png";
					break;
			}
	}
	return 0;//status??
}

typedef struct {
		char* buf;
		int size;
		union {
			int pos;
			char jpeg_unexpected_end;
		}x;
}IODesc;


raw_bitmap *new_raw_bitmap(){
	raw_bitmap * bmp = (raw_bitmap*)malloc(sizeof(raw_bitmap));
	bmp->bitmap = bmp->raster = bmp->palette = (unsigned char*)0;
	bmp->height = bmp->width = bmp->bpp = bmp->pal_bpp = -1;
	bmp->pal_entries = -1;
	return bmp;
}


static void mem_to_png(png_structp png_ptr,
        png_bytep data, png_size_t length)
{
		IODesc *desc=(IODesc*)png_get_io_ptr(png_ptr);
		
		if(desc->x.pos + length >= desc->size) png_error(png_ptr, "Reading past input buffer\n");
		
	memcpy(data, desc->buf + desc->x.pos,length);
	desc->x.pos += length;
}

static void png_warn_still(png_structp png_ptr, png_const_charp msg){}

static void png_err_still(png_structp png_ptr, png_const_charp msg)
{
	longjmp(png_jmpbuf(png_ptr), err_unknown + while_decomp);
}

int png2bitmap(char *inbuf, int insize, raw_bitmap **out)
{
  	png_structp png_ptr;
   	png_infop info_ptr, end_info = NULL;
   	int color_type, grayscale, num_palette, bit_depth;
	png_color *palette;
	png_bytepp row_pointers;
	png_bytep onerow;
	int i;
	IODesc desc;
	raw_bitmap *bmp;

   /* Compare the first 8 bytes of the signature.
      Return with error if they don't match. */

	if((insize < 8) || png_sig_cmp(inbuf, (png_size_t)0, 8))
		return err_bad_data_format + while_decomp;
   
	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
	NULL, NULL, NULL);
	if(!png_ptr) {
			return err_out_of_mem + while_decomp;
	}

	
	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr) {
			return err_out_of_mem + while_decomp;
	}
	if (i = setjmp(png_jmpbuf(png_ptr))) {
	png_destroy_read_struct(&png_ptr, &info_ptr,
	   &end_info);
		return i;
	}

	png_set_error_fn(png_ptr,(void*)0, 
			png_err_still, png_warn_still);	

	desc.buf=inbuf;
	desc.size=insize;
	desc.x.pos=0;

	bmp = new_raw_bitmap();
	*out = bmp;

	png_set_read_fn(png_ptr,(voidp)&desc,mem_to_png);
	
	png_read_info(png_ptr,info_ptr);
	png_get_IHDR(png_ptr,info_ptr,(png_uint_32*)&(bmp->width),
			(png_uint_32*)&(bmp->height),
			&bit_depth,&color_type,(int*)0,(int*)0,(int*)0);

	grayscale = 1;
	
	if (color_type & PNG_COLOR_MASK_PALETTE)
	{
		//TODO better palette->grayscale expansion
		//Detects grayscale image (scans all colors in palette)
/*			png_get_PLTE(png_ptr, info_ptr, &palette, &num_palette);
		for(i=0;i<num_palette;i++)
				if(palette[i].red != palette[i].green ||
							palette[i].green != palette[i].blue)
				{
						grayscale=0;
						break;
				}*/
		png_set_palette_to_rgb(png_ptr);
		// A palette image may or may not have alpha
		// transparency. I'd be safe and remove it anyway.
		png_set_strip_alpha(png_ptr); 

		if(!AllowLookCh){
		//this option needs explicit check for transparency 
			png_bytep trans;
			int num_trans;
			png_color_16p trans_values;
			
			png_get_tRNS(png_ptr, info_ptr, &trans, &num_trans, &trans_values);
			if(num_trans > 0)
				return err_possible_look_change + while_decomp;
		}
	}
	
	if (color_type == PNG_COLOR_TYPE_GRAY || 
				color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
	{
		grayscale=1;	
        if(bit_depth < 8) png_set_gray_1_2_4_to_8(png_ptr);
	}else grayscale=0;
	
	if (bit_depth == 16)
        png_set_strip_16(png_ptr);
	
	 if (color_type & PNG_COLOR_MASK_ALPHA){
		 if(!AllowLookCh) return err_possible_look_change + while_decomp;
	        png_set_strip_alpha(png_ptr);
	 }
	 

	if (grayscale) bmp->bpp = 1;
		else bmp->bpp = 3;

	row_pointers = (png_bytepp)malloc(bmp->height * sizeof(png_bytep));
	
	bmp->bitmap=(unsigned char*)malloc(
			(bmp->width) * (bmp->height) * (bmp->bpp));
	onerow=(png_bytep)bmp->bitmap;

	for (i=0; i < bmp->height; i++) {
	row_pointers[i]=onerow;
		onerow += (bmp->width) * (bmp->bpp);
	}
	png_read_image(png_ptr,row_pointers);

	free(row_pointers);

	png_destroy_read_struct(&png_ptr, &info_ptr,NULL);

	return err_ok;
}

static void png_to_mem(png_structp png_ptr,
        png_bytep data, png_size_t length){
		IODesc *desc=(IODesc*)png_get_io_ptr(png_ptr);
		if(length + desc->x.pos > desc->size)
			png_error(png_ptr, "Writing past output buffer\n");
		
	memcpy(desc->buf + desc->x.pos, data ,length);
	desc->x.pos += length;
}

static void png_flush_mem(png_structp png_ptr){};

static int bitmap2png(raw_bitmap  * bmp, char ** outb, int * outl){
	int i, ctype, bits = 8;
	IODesc desc;
	png_bytepp row_pointers;
	png_bytep onerow;
	png_color_16 transcol;
	char *tr = (char*)0;
	png_infop info_ptr;
	
	png_structp png_ptr = png_create_write_struct
	(PNG_LIBPNG_VER_STRING, NULL,
	NULL, NULL);
	
	if (!png_ptr)
	return (err_unknown + while_compress);

	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr) {
		png_destroy_write_struct(&png_ptr,
		 (png_infopp)NULL);
		return (err_unknown + while_compress);
	}
	
	if (i = setjmp(png_jmpbuf(png_ptr))) {
		png_destroy_write_struct(&png_ptr,
		 &info_ptr);
		return (i);
	}

	png_set_error_fn(png_ptr,(void*)0, 
			png_err_still, png_warn_still);	

	if(bmp->pal_entries > 0){
		png_color *palette;
		
	 	palette = (png_colorp)png_malloc(png_ptr, bmp->pal_entries
		     * sizeof (png_color));
		     
		if(bmp->pal_bpp == 2 || bmp->pal_bpp == 4){
			tr = (char*)malloc(bmp->pal_entries);
		}
		
		palette = malloc(bmp->pal_entries * sizeof(png_color));
		for(i=0;i < bmp->pal_entries;i++){
			if(tr){
				if(0 == bmp->palette[(i+1)*bmp->pal_bpp -1]){
					transcol.index = i;
					tr[i] = 0;
				}else tr[i] = 255;
			}
				
			if(bmp->pal_bpp < 3)
			palette[i].red = palette[i].green = palette[i].blue = bmp->palette[i * bmp->pal_bpp];
			else{
			palette[i].red = bmp->palette[i*bmp->pal_bpp];
			palette[i].green = bmp->palette[i*bmp->pal_bpp + 1];
			palette[i].blue = bmp->palette[i*bmp->pal_bpp + 2];
			}
		}
		ctype = PNG_COLOR_TYPE_PALETTE;
		png_set_PLTE(png_ptr, info_ptr, palette, bmp->pal_entries);

		if (tr) {
			png_set_tRNS(png_ptr, info_ptr, tr, bmp->pal_entries,NULL);
			//png_set_bKGD(png_ptr, info_ptr, &transcol);
			
		}
		if(bmp->pal_entries <= 2) bits = 1;
		else if(bmp->pal_entries <= 4) bits = 2;
		else if(bmp->pal_entries <= 16) bits = 4;
		
		onerow=(png_bytep)bmp->raster;
	}else{
		switch(bmp->bpp){
			case 1: ctype = PNG_COLOR_TYPE_GRAY;
				break;
			case 2: ctype = PNG_COLOR_TYPE_GRAY_ALPHA;
				break;
			case 3: ctype = PNG_COLOR_TYPE_RGB;
				break;
			case 4: ctype = PNG_COLOR_TYPE_RGB_ALPHA;
				break;
		}
/*		color_bpps.red = 8;
		color_bpps.green = 8;
		color_bpps.blue = 8;
		color_bpps.gray = 8;
		color_bpps.alpha = 8;*/

		png_set_filter(png_ptr, 0,PNG_FILTER_SUB);

//		png_set_sBIT(png_ptr, info_ptr, &color_bpps);
		onerow=(png_bytep)bmp->bitmap;
	}

	png_set_IHDR(png_ptr, info_ptr, bmp->width, bmp->height, bits,
			ctype,PNG_INTERLACE_NONE,PNG_COMPRESSION_TYPE_DEFAULT,
			PNG_FILTER_TYPE_DEFAULT);
	
	png_set_compression_level(png_ptr, Z_BEST_COMPRESSION);
	
	row_pointers = (png_bytepp)malloc(bmp->height * sizeof(png_bytep));
	

	for (i=0; i < bmp->height; i++) {
		row_pointers[i]=onerow;
		onerow += (bmp->width) * 
			(ctype == PNG_COLOR_TYPE_PALETTE ? 1 : bmp->bpp);
	}
	
	png_set_rows(png_ptr, info_ptr, row_pointers);
	
	desc.buf = *outb;
	desc.size = *outl;
	desc.x.pos=0;
	png_set_write_fn(png_ptr, (void*)&desc, png_to_mem,
        png_flush_mem);

	png_write_png(png_ptr, info_ptr,PNG_TRANSFORM_PACKING, NULL);

	*outl = desc.x.pos;
	
	return err_ok;
}

static int gif_mem_input(GifFileType *GifFile, GifByteType* data, int bytes)
{
	IODesc *desc = (IODesc*) GifFile->UserData;
	
	int copied;
	
	if(bytes + desc->x.pos >= desc->size) copied = desc->size - desc->x.pos;
	else copied=bytes;
	
	
	memcpy(data,desc->buf + desc->x.pos,copied);

	desc->x.pos+=copied;

	return copied;
}

/**
 * Get transparency color from graphic extension block
 *
 * Return: transparency color or -1
 */
static int getTransparentColor(GifFileType * file)
{
  int i;
  ExtensionBlock * ext = file->SavedImages[0].ExtensionBlocks;
 
  for (i=0; i < file->SavedImages[0].ExtensionBlockCount; i++, ext++) {
  
    if (ext->Function == GRAPHICS_EXT_FUNC_CODE) {
      if (ext->Bytes[0] & 1)/* there is a transparent color */
        return (unsigned char) ext->Bytes[3];/* here it is */
    }
  }

  return -1;
}


int gif2bitmap(char *inbuf, int insize, raw_bitmap **out)
{
	GifFileType *GifFile;
	SavedImage* LastImg = NULL;
	GifImageDesc *LastImgDesc = NULL;
	ColorMapObject *Colors;
	GifColorType * BackGround, *Pixel;
	IODesc desc;
	int i,j,k,c,offset, transColor, colind;
	unsigned char* BufferP, *Raster;
	const int    InterlacedOffset[] = { 0, 4, 2, 1 }; /* The way Interlaced image should. */
	const int    InterlacedJumps[] = { 8, 8, 4, 2 };    /* be read - offsets and jumps... */
	raw_bitmap *bmp;
	unsigned char *MyColorMap = (char*)0;
	int MyColorCount = -1;

	desc.buf=inbuf;
	desc.size=insize;
	desc.x.pos=0;

	if ((GifFile = DGifOpen((void*)&desc, &gif_mem_input)) == NULL) 
    		return( err_unknown + while_decomp);//more possible reasons

	if(DGifSlurp(GifFile)!=GIF_OK)
	{
			fprintf(stderr,"GIF reading failed\n");
			return(err_bad_data_format + while_decomp);
	}
	bmp = new_raw_bitmap();
	*out = bmp;
	bmp->width=GifFile->SWidth;
	bmp->height=GifFile->SHeight;
	transColor = getTransparentColor(GifFile);

	//We're interested in last frame. Searching for last complete redraw
	//of image.
	i = GifFile->ImageCount - 1;
	if(i > 0 && !AllowLookCh) return err_possible_look_change + while_decomp;
	for(; i>=0; i--)
		if(GifFile->SavedImages[i].ImageDesc.Left == 0 &&
			GifFile->SavedImages[i].ImageDesc.Top == 0 &&
			GifFile->SavedImages[i].ImageDesc.Width == bmp->width &&
			GifFile->SavedImages[i].ImageDesc.Height == bmp->height
			)
				LastImg = GifFile->SavedImages + i;

	if(LastImg) {
		
		Raster = (unsigned char*)malloc(bmp->width*bmp->height);
		memcpy(Raster, LastImg->RasterBits, bmp->width*bmp->height);
		
		LastImgDesc = &LastImg->ImageDesc;
	
		Colors = (LastImgDesc->ColorMap
			? LastImgDesc->ColorMap
			: GifFile->SColorMap);

	}else{

		 if(!AllowLookCh) return err_possible_look_change + while_decomp;
		//no complete image found - Make overlay of all images
		Raster = (unsigned char*)malloc(bmp->width*bmp->height);
		
		for(i=0;i<bmp->height;i++)
			for(j=0;j<bmp->width;j++)
				Raster[i*bmp->width + j] = GifFile->SBackGroundColor;
		
		for(i=0;i<bmp->height;i++)
			for(j=0;j<bmp->width;j++)
				for(k = 0;k < GifFile->ImageCount; k++){
					int left = GifFile->SavedImages[k].ImageDesc.Left;
					int right = left + GifFile->SavedImages[k].ImageDesc.Width;
					int top = GifFile->SavedImages[i].ImageDesc.Top;
					int bottom = top + GifFile->SavedImages[i].ImageDesc.Height == bmp->height;

					if(i >= top && i <= bottom && j >= left && j <= right){
						Raster[(i - top)*bmp->width + (j-left)] =
							GifFile->SavedImages[k].RasterBits[(i - top)*(left-right) + (j-left)];
						break;
					}
					
				}
		Colors = GifFile->SColorMap;
	}
	
	//Often there are unused colors in the palette.
	MyColorMap = (unsigned char*)malloc(Colors->ColorCount);
	for(i=0; i< Colors->ColorCount;i++)
		MyColorMap[i] = 0;
	
	MyColorCount = 0;
	for(i=0; i< bmp->width*bmp->height;i++){
		if(!MyColorMap[Raster[i]]){
			MyColorCount++; 
			//We're numbering the colors here from 1 upwards. But in 
			//the final raster they'd be from 0.
			MyColorMap[Raster[i]] = MyColorCount;
			if((255 == MyColorCount) || (MyColorCount == Colors->ColorCount)){
				free(MyColorMap);//we're using full palette
				MyColorMap = (char*)0;
				MyColorCount = Colors->ColorCount;
				break;
			}
		}
	}
	
	bmp->bpp=1;
	for(i = 0; i < Colors->ColorCount; i++)
	{
		if((i != transColor) && (!MyColorMap || MyColorMap[i]) &&
			(( Colors->Colors[i].Red !=  Colors->Colors[i].Green ) ||
			( Colors->Colors[i].Green != Colors->Colors[i].Blue)))
		{
			bmp->bpp=3;
			break;
		}
	}
	if (MyColorMap && !MyColorMap[transColor])
			transColor = -1;

	if (transColor >= 0 ){
		if(!AllowLookCh) return err_possible_look_change + while_decomp;
		bmp->bpp++; //add alpha channel
	}
	
	assert(bmp->bpp > 0 && bmp->bpp <= 4);
	
	bmp->pal_entries = MyColorCount;
	bmp->palette = (unsigned char*)malloc(bmp->pal_entries*bmp->bpp);
	
	BufferP = bmp->palette;
	for(i = 0; i < Colors->ColorCount; i++){
		if(MyColorMap) {
			if(!MyColorMap[i]) continue;
			BufferP = bmp->palette + bmp->bpp*(MyColorMap[i] - 1);
		}
		Pixel = Colors->Colors + i;
		for(k=0;k<bmp->bpp;k++)
			switch(k){
				case 0: *BufferP++ = Pixel->Red;
					break;
				case 1: if( bmp->bpp > 2)
						*BufferP++ = Pixel->Green;
					else 
						*BufferP++ = 
						( i == transColor ?
						0 : 255);

					break;
				case 2: *BufferP++ = Pixel->Blue;
					break;
				case 3: *BufferP++ = 
					( i == transColor ?
					0 : 255);
					break;
			}

	}
	
	
	if(GifFile->Image.Interlace){
		bmp->raster=(unsigned char*)malloc(bmp->width * bmp->height);
		BufferP=bmp->raster;
		offset = -1;
			
	    for (c = 0; c < 4; c++)
		for (i = InterlacedOffset[c]; i < bmp->height;
					 i += InterlacedJumps[c]){
			BufferP = bmp->raster + i*bmp->width;
			for(j=0; j < bmp->width; j++){
				offset++;
				if(MyColorMap)
					*BufferP++ = MyColorMap[Raster[offset]] - 1;
				else
					*BufferP++ = Raster[offset];
			}
		}
	    free(Raster);

	}else{
		if(MyColorMap) for(i =0; i< bmp->width*bmp->height; i++){
			Raster[i]=MyColorMap[Raster[i]] - 1;
		}
		
		bmp->raster = Raster;
	}
	bmp->pal_bpp = bmp->bpp;
	
	DGifCloseFile(GifFile);
	return err_ok;
}

IODesc * get_src (IODesc* ptr)
{
	static IODesc *myptr;

	if (ptr) myptr = ptr;
	return myptr;
}

static void jpeg_src_noop(j_decompress_ptr dinfo){}

static void jpeg_src_init(j_decompress_ptr dinfo)
{
	IODesc * desc = get_src(NULL);

	dinfo-> src ->next_input_byte = desc->buf;
	dinfo-> src -> bytes_in_buffer = desc->size;
}

static boolean jpeg_src_fill_input(j_decompress_ptr dinfo)
{
	return FALSE;
}

static void jpeg_src_skip_input (j_decompress_ptr dinfo, long num_bytes)
{
	IODesc * desc = get_src(NULL);

	if ((char*)(dinfo->src->next_input_byte + num_bytes) > desc->buf + desc->size || 
		(char*)dinfo->src->next_input_byte == &desc->x.jpeg_unexpected_end )
		{ //silly behavior - recommended by jpeglib docs
			dinfo->src->next_input_byte = &desc->x.jpeg_unexpected_end;
			dinfo->src->bytes_in_buffer = 1; 
		}
		else
		{
			dinfo->src->next_input_byte += num_bytes;
			dinfo->src->bytes_in_buffer -= num_bytes;
		}
}

jmp_buf errjmp;

static void jpeg_error_exit (j_common_ptr cinfo)
{
	longjmp(errjmp,1);
}

static void jpeg_emit_message (j_common_ptr cinfo, int msg_level)
{}



int jpg2bitmap(char *inbuf, int insize, raw_bitmap **out)
{
		
        struct jpeg_decompress_struct dinfo;
	char *BufferP;
        struct jpeg_error_mgr jerr;
	struct jpeg_source_mgr jsrc;
	IODesc desc;
	int imgsize,row_stride;
	raw_bitmap *bmp;

	desc.buf = inbuf;
	desc.size = insize;
	desc.x.jpeg_unexpected_end = JPEG_EOI;

	get_src(&desc);

	dinfo.err = jpeg_std_error(&jerr);
	dinfo.err->error_exit = jpeg_error_exit;
	dinfo.err->emit_message = jpeg_emit_message;
	
        jpeg_create_decompress(&dinfo);

	jsrc.init_source = jpeg_src_init;
	jsrc.term_source = jpeg_src_noop; 
	jsrc.fill_input_buffer = jpeg_src_fill_input;
	jsrc.skip_input_data = jpeg_src_skip_input;
	dinfo.src = &jsrc;
	
	if (setjmp(errjmp))
			return err_unknown + while_decomp;

	bmp = new_raw_bitmap();
	*out = bmp;
	jpeg_read_header(&dinfo, TRUE);
	jpeg_start_decompress(&dinfo);
	
	imgsize=dinfo.output_width*dinfo.output_height*dinfo.output_components;
	row_stride=dinfo.output_width*dinfo.output_components;
	
	bmp->width=dinfo.output_width;
	bmp->height=dinfo.output_height;
	bmp->bpp=dinfo.output_components;
	
	bmp->bitmap = (unsigned char*)malloc (imgsize);
	BufferP=bmp->bitmap;
	while (dinfo.output_scanline < dinfo.output_height)
	{
			jpeg_read_scanlines(&dinfo,(JSAMPARRAY)(&BufferP),1);
			BufferP+=row_stride;
	}
	
	jpeg_finish_decompress(&dinfo);
	jpeg_destroy_decompress(&dinfo);
	return err_ok;
}

static void jpeg_dest_init(j_compress_ptr cinfo)
{
	outbuf = cinfo->dest->next_output_byte = malloc(outlen);
	cinfo->dest->free_in_buffer=outlen;
}

static boolean jpeg_dest_empty_output_buffer()
{
		return FALSE; //output file larger than original..
}

static void jpeg_dest_term(j_compress_ptr cinfo)
{
		outlen = outlen - cinfo->dest->free_in_buffer;
}

int bitmap2jpg(raw_bitmap  * bmp, int quality, char ** outb, int * outl)
{
	struct jpeg_compress_struct cinfo;
        struct jpeg_error_mgr jerr;
	struct jpeg_destination_mgr memdest;
	int row_stride = bmp->width*bmp->bpp;
	int i,j;
	unsigned char *BufferP, *noA;

	cinfo.err = jpeg_std_error(&jerr);
	cinfo.err->error_exit = jpeg_error_exit;
	cinfo.err->emit_message = jpeg_emit_message;


	//translate i -> error code ?
	if(i = setjmp(errjmp))
		return err_unknown + while_compress;

	jpeg_create_compress(&cinfo);
	
	memdest.init_destination = jpeg_dest_init;
	memdest.empty_output_buffer = jpeg_dest_empty_output_buffer;
	memdest.term_destination = jpeg_dest_term;
	
	cinfo.dest = &memdest;
	
	cinfo.image_width = bmp->width; 	/* image width and height, in pixels */
	cinfo.image_height = bmp->height;

	if(bmp->bpp <= 2){
		cinfo.input_components = 1;
		cinfo.in_color_space = JCS_GRAYSCALE;
	}else{
		cinfo.input_components = 3;	
		cinfo.in_color_space = JCS_RGB; /* colorspace of input image */
	}
		
	jpeg_set_defaults(&cinfo);
	
	if(quality > 100){
		quality -=128;
		cinfo.smoothing_factor = 10;
	}
	
	cinfo.optimize_coding = TRUE;
	jpeg_set_quality(&cinfo, quality, FALSE );

	jpeg_start_compress(&cinfo, TRUE);
	BufferP = bmp->bitmap;

	if(bmp->bpp != cinfo.input_components)
		noA = (unsigned char *)malloc(bmp->width * cinfo.input_components);
	
	for(i=0;(i == cinfo.next_scanline) &&
			(cinfo.next_scanline < cinfo.image_height);i++) {
		//remove alpha channel if any
		if(bmp->bpp != cinfo.input_components){
			unsigned char * p = noA;
			for(j=0;j < bmp->width * bmp->bpp;j++)
				if(bmp->bpp - 1 != (j % bmp->bpp)){
					*p = BufferP[j];
					p++;
				}
		}else noA = BufferP;
		
		jpeg_write_scanlines(&cinfo,(JSAMPARRAY)(&noA),1);
		BufferP+=row_stride;
	}
		
	if (i < cinfo.image_height) {
		jpeg_abort_compress(&cinfo);
		outlen = -1;
		free(outbuf);
		return err_out_of_mem + while_compress;
	}

	jpeg_finish_compress(&cinfo);
	jpeg_destroy_compress(&cinfo);
	*outb = outbuf;
	*outl = outlen;
	return err_ok;
}

//what about gamma?
int rgb2gray(raw_bitmap * bmp) {
	int i, Y, size = bmp->height * bmp->width;
	
	if(bmp->bpp != 3 || !bmp->bitmap) return while_transform + err_bad_data_format;

	bmp->pal_bpp = bmp->bpp;
	
	for(i=0; i < size;i++){
	//approx. formula from libpng
	Y = (6969 * bmp->bitmap[3*i] + 
			23434 * bmp->bitmap[3*i+1] + 
			2365 * bmp->bitmap[3*i+2] )/32768;
		if(Y > 255)
			bmp->bitmap[i]=255;
		else
			bmp->bitmap[i]=Y;
	}
	bmp->bpp = 1;
	return err_ok;
}

static int depalettize(raw_bitmap * bmp){
	
	int i, j, k, size = bmp->height * bmp->width;
	unsigned char *Buffer, *BufferP;
	
	if(bmp->pal_entries <= 0) return while_transform + err_bad_data_format;

	Buffer = BufferP = (unsigned char*)malloc(size*bmp->pal_bpp);
	
	for(i=0;i<size;i++)
		for(j=0;j<bmp->pal_bpp;j++)
			*BufferP++ = bmp->palette[bmp->raster[i]*bmp->pal_bpp + j];
	
	bmp->bitmap = Buffer;
	
	return err_ok;	
}

#ifdef JP2K
int bitmap2jp2(raw_bitmap * bmp, float rate, char ** outb, int * outl){
	jas_image_t *jimg;
	jas_image_cmptparm_t cmptparms[4];
	jas_image_cmptparm_t *cmptparm;
	uint_fast16_t cmptno;
	int x,y,z;
	jas_stream_t *js;
	char mode[30], *cmode;

	if(jas_init()) return 5;
	
	if(bmp->pal_entries > 0) depalettize(bmp);
	
	//Prepare components.
	for (cmptno = 0, cmptparm = cmptparms; cmptno < bmp->bpp; ++cmptno,
	  ++cmptparm) {
		cmptparm->tlx = 0;
		cmptparm->tly = 0;
		cmptparm->hstep = 1;
		cmptparm->vstep = 1;
		cmptparm->width = bmp->width;
		cmptparm->height = bmp->height;
		cmptparm->prec = 8;
		cmptparm->sgnd = false;
	}

	if (!(jimg = jas_image_create(bmp->bpp, cmptparms,
	  JAS_CLRSPC_UNKNOWN))) {
		return 0;
	}
	switch (bmp->bpp){
		case 2:
		jas_image_setcmpttype(jimg, 1,
				JAS_IMAGE_CT_OPACITY);
		case 1:
		jas_image_setclrspc(jimg, JAS_CLRSPC_SGRAY);
		jas_image_setcmpttype(jimg, 0,
		  JAS_IMAGE_CT_COLOR(JAS_CLRSPC_CHANIND_GRAY_Y));
		break;

		case 4:
		jas_image_setcmpttype(jimg, 3,
				JAS_IMAGE_CT_OPACITY);
		case 3:
		jas_image_setclrspc(jimg, JAS_CLRSPC_SRGB);
		jas_image_setcmpttype(jimg, 0,
		  JAS_IMAGE_CT_COLOR(JAS_CLRSPC_CHANIND_RGB_R));
		jas_image_setcmpttype(jimg, 1,
		  JAS_IMAGE_CT_COLOR(JAS_CLRSPC_CHANIND_RGB_G));
		jas_image_setcmpttype(jimg, 2,
		  JAS_IMAGE_CT_COLOR(JAS_CLRSPC_CHANIND_RGB_B));
		break;
	}

	for(y=0;y<bmp->height;y++)
		for(x=0;x<bmp->width;x++)
			for(z=0;z<bmp->bpp;z++){
			//write pixel by pixel...slow?
				jas_image_writecmptsample(jimg, z, x, y,	
				bmp->bitmap[bmp->bpp*((y*bmp->width)+x)+z]);
		}
	
	js = jas_stream_memopen(0, 0);
	if(!js) return 0;
	
	if(rate > 0.9999 ){
		cmode= "int"; 
		rate = 1.0;
	}
	else cmode = "real";

	snprintf(mode, sizeof(mode), "mode=%s rate=%f", cmode, rate);
	if( jp2_encode(jimg, js, mode)) return 0;

	jas_stream_flush(js);
	//May not be future compatible!
	jas_stream_memobj_t * jmem = (jas_stream_memobj_t *)js->obj_;
	*outl = jmem->len_;
	*outb = jmem->buf_;
	jmem->myalloc_ = 0; //don't delete buf_ on closing
	jas_stream_close(js);//close

	/*	
	outlen = jas_stream_tell(js);
	outbuf = malloc(outlen);
	jas_stream_rewind(js);
	jas_stream_read(js, outbuf, outlen);
	jas_stream_close(js);*/
	
	return err_ok;
}
#endif
