/* log.c - part of ziproxy package
 *
 * Copyright (c)2003-2004 Juraj Variny<variny@naex.sk>
 * Copyright (c)2005-2006 Daniel Mealha Cabrita
 *
 * Released subject to GNU General Public License.
 *
 * Convenient logging functions.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <sys/time.h>
#include "cfgfile.h"
#include "http.h"

#define LOGGING (LogFileName || LogPipe)

static FILE *logfile;
static int logpid;

static struct timeval newtime, oldtime;
static long timeval_subtract_us ( struct timeval *x, struct timeval *y);
static void goodbye(void);

static FILE *accesslog_file = NULL;

void init_logging(){
	    time_t tm;
	if(!LOGGING) return;
	
	logpid = getpid();
	time(&tm);
    //open logging file
	if(LogPipe)
		logfile = stderr;
	else if(LogFileName) {
		struct tm *btime;
		time_t cas;
		char logname[50];

		time(&cas);
		btime = localtime(&cas);
		strftime(logname,sizeof(logname),LogFileName,btime);
		logfile = fopen(logname, "a");
	}	
	//set line buffered mode for logfile
	setvbuf(logfile,NULL,_IOLBF,BUFSIZ);
	//setlinebuf(logfile);
	
	fprintf(logfile, "[%d] Starting - %s", logpid, ctime(&tm));

	gettimeofday(&oldtime,NULL);
	atexit(&goodbye);
}


int logprintf(char *fmt, ...){
	va_list ap;
	int i;
	if(!LOGGING) return 0;
	va_start(ap, fmt);
	fprintf(logfile,"[%d] ", logpid);
	i = vfprintf(logfile, fmt, ap);
	va_end(ap);
	return i;
}

void logputs(char *str){
	char ch, *fmt;
	if(!LOGGING) return;
//	check_trim(str);//strip any trailing newlines
	fprintf(logfile,"[%d] %s\n", logpid, str);
}

void logputs_hdr(char *str){
	if(!LOGGING) return;
	fprintf(logfile,"[%d]   %s\n", logpid, str);
}

void logdifftime(char * activity){
	if(!LOGGING) return;
	gettimeofday(&newtime,NULL);
	fprintf(logfile, "[%d] %s took %ld us. \n", 
			logpid, activity,
			timeval_subtract_us(&newtime, &oldtime));
	gettimeofday(&oldtime,NULL);

}

static long timeval_subtract_us (struct timeval *x, struct timeval *y){
	struct timeval result;
	
        /* Perform the carry for the later subtraction by updating Y. */
        if (x->tv_usec < y->tv_usec) {
          int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
          y->tv_usec -= 1000000 * nsec;
          y->tv_sec += nsec;
        }
        if (x->tv_usec - y->tv_usec > 1000000) {
          int nsec = (x->tv_usec - y->tv_usec) / 1000000;
          y->tv_usec += 1000000 * nsec;
          y->tv_sec -= nsec;
        }
      
        /* Compute the time remaining to wait.
           `tv_usec' is certainly positive. */
        result.tv_sec = x->tv_sec - y->tv_sec;
        result.tv_usec = x->tv_usec - y->tv_usec;
      
        return 1000000*result.tv_sec + result.tv_usec;
}


static void goodbye(void){
	logputs("Finished.");
}

/* returns: 0=ok 1=error */
int init_access_log (void) {
	if (AccessLogFileName != NULL) {
		accesslog_file = fopen (AccessLogFileName, "a");;
		if (accesslog_file != NULL) {
			// set line buffered mode for access logfile
			setvbuf (accesslog_file, NULL, _IOLBF, BUFSIZ);
	
			return (0);
		}
		return (1);	
	}
	return (0);
}

/*
 *  there's no "close_access_log" function
 * when a process finishes, the file handler is automatically
 * closed by the OS
 */

void access_log (http_headers *client_hdr, int inlen, int outlen, const char *client_addr, struct timeval *giventime) {
	char flags_str[10];
	int has_client_addr = 0;
	struct timeval currtime;
	int time_spent_msec;

	if (accesslog_file == NULL)
		return;
	
	gettimeofday (&currtime, NULL);
	time_spent_msec = timeval_subtract_us (&currtime, giventime) / 1000;
	
	if (client_addr != NULL)
		has_client_addr = 1;
	
	flags_str[0]='\0';
	// Transparent or normal Proxy?
	if (client_hdr->flags & H_TRANSP_PROXY_REQUEST) {
		strcat (flags_str, "T");
	} else {
		strcat (flags_str, "P");
	}
	// SSL?
	if (client_hdr->flags & H_USE_SSL)
		strcat (flags_str, "S");
	
	fprintf (accesslog_file, "%d.%03d %6d %15s %2s %6d %6d %s %s\n", currtime.tv_sec, currtime.tv_usec / 1000, time_spent_msec, has_client_addr?client_addr:"?", flags_str, inlen, outlen, client_hdr->method, client_hdr->url);
}

